#!/usr/bin/python3
#  coding: utf-8
#
#   createAbaBeamVtu.py
#
#       calculixのbeam要素（B31,B32,B31R,B32R）の形状と変位をvtuに変換する
#
#   21/02/06    新規作成
#      02/12    バネ要素とバネ追加の為に追加した節点を変換しない様に修正
#

import os, sys
import pyFistr
import convertMesh as cm


convAbaTypeToVtkDict = {
    #abaType  vtkType
    "B31":      "3",
    "B31R":     "3",
    "B32":      "21",
    "B32R":     "21"
    }

#
#  createNodeElementDict
#------------------------
def createNodeElementDict(meshHeaderData):
    """ nodeDict, elementDictを作成する"""
    nodeDict = {}
    elementDict = {}
    for header, data in meshHeaderData:
        words = cm.deleteSp(header).split(",")
        if words[0] == "*NODE":
            for nodeLine in data:
                nodeNo = nodeLine[0]
                loc = nodeLine[1:]
                nodeDict[nodeNo] = {"loc":loc, "elms":set([])}
    for header, data in meshHeaderData:
        words = cm.deleteSp(header).split(",")
        if words[0] == "*ELEMENT":
            elmType = pyFistr.getValName(header, "TYPE")
            for elmLine in data:
                elmNo = elmLine[0]
                nodes = elmLine[1:]
                elementDict[elmNo] = {"nodes":nodes, "type":elmType}
                for nodeNo in nodes:
                    nodeDict[nodeNo]["elms"].add(elmNo)
    return nodeDict, elementDict

#
#  getNewNumberForVtk
#---------------------
def getNewNumberForVtk(neDicts):
    """ vtk用のNoを辞書内に取得する"""
    nodeDict, elementDict = neDicts
    nodes = list(nodeDict.keys())
    nodes.sort()
    for i in range(len(nodes)):
        nodeNo = nodes[i]
        nodeDict[nodeNo]["newNo"] = i
    elms = list(elementDict.keys())
    elms.sort()
    for i in range(len(elms)):
        elmNo = elms[i]
        elementDict[elmNo]["newNo"] = i
    return nodeDict, elementDict

#
#  getLatestDispTable
#---------------------
def getLatestDispTable(lines):
    ip = -1
    for i in range(len(lines)-1, -1, -1):
        line = lines[i]
        if line.find("displacements") >= 0:
            ip = i
            break
    print("displacementを取得")
    print("  " + line)
    disps = []
    for i in range(ip+1, len(lines), 1):
        line = lines[i]
        words = line.split()
        if len(words) == 4:
            try:
                nodeNo = int(words[0])
                disp = list(map(float, words[1:]))
                dispLine = [nodeNo] + disp
                disps.append(dispLine)
            except:
                break
        elif len(words) > 4:
            break
    return disps

#
#  setDispInNodeDict
#---------------------
def setDispInNodeDict(tableData, neDicts):
    nodeDict, elementDict = neDicts
    for line in tableData:
        nodeNo = line[0]
        disp = line[1:]
        nodeDict[nodeNo]["disp"] = disp
    return nodeDict, elementDict

#
#  getElmTypeDict
#-----------------
def getElmTypeDict(neDicts):
    nodeDict, elementDict = neDicts
    elmTypeDict = {}
    for elmNo in elementDict.keys():
        elmType = elementDict[elmNo]["type"]
        if elmType in elmTypeDict.keys():
            elmTypeDict[elmType].append(elmNo)
        else:
            elmTypeDict[elmType] = []
    return elmTypeDict

#
#  createVtkLines
#------------------
def createVtkLines(neDicts):
    nodeDict, elementDict = neDicts
    #headerを作成
    lines = ['<?xml version="1.0"?>\n']
    lines += ['<VTKFile type="UnstructuredGrid" version="1.0">\n']
    lines += ["<UnstructuredGrid>\n"]
    #必要なnodeNoを取得（バネ要素を除く）
    allNodes = list(nodeDict.keys())
    nodes = []
    for nodeNo in allNodes:
        #変位を読み取っていないnode（バネ）を除く
        if "disp" in nodeDict[nodeNo].keys():
            nodes.append(nodeNo)
    nodes.sort()
    #必要な要素を取得（バネ要素を除く）
    allElms = list(elementDict.keys())
    elms = []
    for elm in allElms:
        #typeを確認し、辞書以外の要素は除く
        if elementDict[elm]["type"] in convAbaTypeToVtkDict.keys():
            elms.append(elm)
    elms.sort()
    lines += ['<Piece NumberOfPoints="' + str(len(nodes)) + '" NumberOfCells="' + str(len(elms)) + '">\n']
    #NODEを作成
    lines += ["<Points>\n"]
    lines += ['<DataArray type="Float32" NumberOfComponents="3" format="ascii">\n']
    for nodeNo in nodes:
        loc = nodeDict[nodeNo]["loc"]
        line = " ".join(list(map(str, loc))) + "\n"
        lines += [line]
    lines += ['</DataArray>\n']
    lines += ['</Points>\n']
    #elementを作成
    lines += ['<Cells>\n']
    lines += ['<DataArray type="Int32" Name="connectivity" format="ascii">\n']
    #  connectivity
    for elmNo in elms:
        abaNodes = elementDict[elmNo]["nodes"]
        newNodes = []
        for nodeNo in abaNodes:
            newNo = nodeDict[nodeNo]["newNo"]
            newNodes.append(newNo)
        line = " ".join(list(map(str, newNodes))) + "\n"
        lines += [line]
    lines += ['</DataArray>\n']
    #  offsets
    lines += ['<DataArray type="Int32" Name="offsets" format="ascii">\n']
    offsets = []
    offset = 0
    for elmNo in elms:
        abaNodes = elementDict[elmNo]["nodes"]
        offset += len(abaNodes)
        offsets.append(offset)
    line = " ".join(list(map(str, offsets))) + "\n"
    lines += [line]
    lines += ['</DataArray>\n']
    #  types
    lines += ['<DataArray type="UInt8" Name="types" format="ascii">\n']
    types = []
    for elmNo in elms:
        elmType = elementDict[elmNo]["type"]
        vtkType = convAbaTypeToVtkDict[elmType]
        types.append(vtkType)
    line = " ".join(types) + "\n"
    lines += [line]
    lines += ['</DataArray>\n']
    lines += ['</Cells>\n']
    #disp
    lines += ['<PointData>\n']
    lines += ['<DataArray type="Float32" Name="DISPLACEMENT" NumberOfComponents="3" format="ascii">\n']
    for nodeNo in nodes:
        disp = nodeDict[nodeNo]["disp"]
        line = " ".join(list(map(str, disp))) + "\n"
        lines += [line]
    lines += ['</DataArray>\n']
    lines += ['</PointData>\n']
    lines += ['</Piece>\n']
    lines += ['</UnstructuredGrid>\n']
    lines += ['</VTKFile>\n']
    return lines

#-----------------
#  createBeamVtu
#-----------------
def createBeamVtu(fileHeader):
    """ calculixのbeam要素とその変位をvtu形式に変換する"""
    inpFile = fileHeader + ".inp"
    f = open(inpFile); lines = f.readlines(); f.close()
    meshHeaderData = cm.getAbaHeaderNumData(lines)
    neDicts = createNodeElementDict(meshHeaderData)
    neDicts = getNewNumberForVtk(neDicts)
    resFile = fileHeader + ".dat"
    f = open(resFile); lines = f.readlines(); f.close()
    dispTable = getLatestDispTable(lines)
    neDicts = setDispInNodeDict(dispTable, neDicts)
    lines = createVtkLines(neDicts)
    vtuFile = fileHeader + ".vtu"
    f = open(vtuFile, "w"); f.writelines(lines); f.close()
    print("calculixの結果をvtuに変換しました。")

if __name__ == "__main__":
    fileHeader = sys.argv[1]
    createBeamVtu(fileHeader)
