#!/usr/bin/python3
#
#   fistr2abaqusForBeam.py
#
#       fistr形式のmsh、cntファイルをabaqus形式のinpファイルに変換する
#       beam要素とEQIATION
#
#   21/02/04    新規作成
#      02/07    createLines:座標値の出力を「16.9e」で指定。calculix側で
#               エラー発生する場合がある。
#      02/11    !SPRINGの変換を追加
#

import sys, os
import geometryFunc as geo
import convertMesh as cm
import pyFistr


fistrToAbaqus_typeDict = {
    "611": "B31R",
    "612": "B32R"
    }

#変換flag
#   "change2nd":  2次要素に変換する
#   "noChange":   そのまま変換する
meshFlag = "change2nd"


#
#  changeOrder
def changeOrder(elmType, nodes):
    """ typeに応じて節点順を修正"""
    if elmType == "611":
        return nodes
    elif elmType == "612":
        return nodes
    else:
        return []

#
#  createNodeDict
def createNodeDict(meshHeaderData):
    """ nodeDictを作成する"""
    nodeDict = {}
    for header, data in meshHeaderData:
        words = pyFistr.deleteSp(header).split(",")
        if words[0] == "!NODE":
            for node in data:
                nodeNo = node[0]
                loc = node[1:]
                nodeDict[nodeNo] = {"loc":loc, "elms":[]}
    return nodeDict

#
#  getNodeCard
def getNodeCard(meshHeaderData):
    addHeaderData = []
    for headerData in meshHeaderData:
        header = headerData[0]
        words = pyFistr.deleteSp(header).split(",")
        if words[0] == "!NODE":
            newHeader = "*NODE, NSET=NALL\n"
            addHeaderData.append([newHeader, headerData[1]])
    return addHeaderData

#
#  getElementCard
def getElementCard(meshHeaderData, nodeDict):
    addHeaderData = []
    addNodeData = []
    elementDict = {}
    maxNo = max(list(nodeDict.keys()))
    for headerData in meshHeaderData:
        header = headerData[0]
        data = headerData[1]
        words = pyFistr.deleteSp(header).split(",")
        if words[0] == "!ELEMENT":
            elmType = pyFistr.getValName(header, "TYPE")
            if elmType in fistrToAbaqus_typeDict.keys():
                if elmType == "611" and meshFlag == "change2nd":
                    #2次要素に変換する
                    abaType = fistrToAbaqus_typeDict["612"]
                    egrp = pyFistr.getValName(header, "EGRP")
                    if egrp == "":
                        egrp = abaType
                    newHeader = "*ELEMENT, TYPE=" + abaType + ", ELSET=" + egrp + "\n"
                    for i in range(len(data)):
                        elmData = data[i]
                        elmNo = elmData[0]
                        nodes = elmData[1:]
                        elementDict[elmNo] = {"nodes":nodes, "type":elmType}
                        for nodeNo in nodes:
                            nodeDict[nodeNo]["elms"].append(elmNo)
                        loc1 = nodeDict[nodes[0]]["loc"]
                        loc2 = nodeDict[nodes[1]]["loc"]
                        addPointLoc = geo.midPoint(loc1, loc2)
                        maxNo += 1
                        newNo = maxNo
                        nodeDict[newNo] = {"loc":addPointLoc, "elms":[elmNo]}
                        newNodeData = [newNo] + addPointLoc
                        addNodeData.append(newNodeData)
                        newNodes = [nodes[0], newNo, nodes[1]]
                        nodes = changeOrder("612", newNodes)
                        elmData = [elmData[0]] + nodes
                        data[i] = elmData
                    addHeaderData.append([newHeader, data])
                elif elmType == "612" or (elmType == "611" and meshFlag != "change2nd"):
                    #そのまま変換
                    abaType = fistrToAbaqus_typeDict[elmType]
                    egrp = pyFistr.getValName(header, "EGRP")
                    if egrp == "":
                        egrp = abaType
                    newHeader = "*ELEMENT, TYPE=" + abaType + ", ELSET=" + egrp + "\n"
                    for i in range(len(data)):
                        elmData = data[i]
                        elmNo = elmData[0]
                        nodes = elmData[1:]
                        elementDict[elmNo] = {"nodes":nodes, "type":elmType}
                        nodes = changeOrder(elmType, nodes)                    
                        elmData = [elmData[0]] + nodes
                        data[i] = elmData
                    addHeaderData.append([newHeader, data])
    if len(addNodeData) > 0:
        header = "*NODE, NSET=NALL\n"
        addNodeHeaderData = [[header, addNodeData]]
    else:
        addNodeHeaderData = []
    return addHeaderData, addNodeHeaderData, nodeDict, elementDict

#
#  getNgrpCard
def getNgrpCard(meshHeaderData):
    addHeaderData = []
    for header, data in meshHeaderData:
        words = pyFistr.deleteSp(header).split(",")
        if words[0] == "!NGROUP":
            ngrp = pyFistr.getValName(header, "NGRP")
            newHeader = "*NSET, NSET=" + ngrp + "\n"
            addHeaderData.append([newHeader, data])
    return addHeaderData

#
#  getSectionCard
def getSectionCard(meshHeaderData, elementDict):
    """ sectionCardを作成"""

    def getEgrpElementType(meshHeaderData, egrp):
        elmType = ""
        for header, data in meshHeaderData:
            words = pyFistr.deleteSp(header).split(",")
            if words[0] == "!ELEMENT":
                if pyFistr.getValName(header, "EGRP") == egrp:
                    elmType = pyFistr.getValName(header, "TYPE")
                    break
            if words[0] == "!EGROUP":
                if pyFistr.getValName(header, "EGRP") == egrp:
                    elmNo = data[0]
                    elmType = elementDict[elmNo]["type"]
                    break
        return elmType

    addHeaderData = []
    for header, data in meshHeaderData:
        words = pyFistr.deleteSp(header).split(",")
        if words[0] == "!SECTION":
            secType = pyFistr.getValName(header, "TYPE")
            egrp = pyFistr.getValName(header, "EGRP")
            matName = pyFistr.getValName(header, "MATERIAL")
            elmType = getEgrpElementType(meshHeaderData, egrp)
            if meshFlag == "change2nd":
                newHeader = "*BEAM GENERAL SECTION, ELSET=" + egrp + ", MATERIAL=" + matName + ", SECTION=PIPE\n"
                comm    = "** phi, t\n"
                recData = "0.100, 0.004\n"
            else:
                newHeader = "*BEAM SECTION, ELSET=" + egrp + ", MATERIAL=" + matName + ", SECTION=RECT\n"
                comm    = "** width, height\n"
                recData = "0.100, 0.100\n"
            ws = pyFistr.deleteSp(data[0]).split(",")
            dirData = ", ".join(ws[:3]) + "\n"
            newData = []
            newData.append(comm)
            newData.append(recData)
            newData.append(dirData)
            addHeaderData.append([newHeader, newData])
    return addHeaderData

#
#  getEquationCard
def getEquationCard(meshHeaderData):
    """ equationCardを作成"""
    addHeaderData = []
    for header, data in meshHeaderData:
        words = pyFistr.deleteSp(header).split(",")
        if words[0] == "!EQUATION":
            newData = []
            for i in range(0, len(data), 2):
                line1 = data[i]
                line2 = data[i+1]
                ws = pyFistr.deleteSp(line1).split(",")
                line1 = ws[0] + "\n"
                newData.append(line1)
                newData.append(line2)
            newHeader = "*EQUATION\n"
            addHeaderData.append([newHeader, newData])
    return addHeaderData

#
#  getSpringCard
def getSpringCard(contHeaderData, meshHeaderData, nodeDict, elementDict):
    """ springCardを作成。
    xyz各方向に長さ1.0のバネを追加する。"""
    
    def getNgrpData(meshHeaderData, ngrpName):
        ngrpData = []
        for header, data in meshHeaderData:
            words = pyFistr.deleteSp(header).split(",")
            if words[0] == "!NGROUP":
                ngrp = pyFistr.getValName(header, "NGRP")
                if ngrp == ngrpName:
                    ngrpData += data
            elif words[0] == "!NODE":
                ngrp = pyFistr.getValName(header, "NGRP")
                if ngrp == ngrpName:
                    nodeNos = list(map(lambda x: x[0], data))
                    ngrpData += nodeNos
        return ngrpData

    spHeaderData = []
    elsetHeaderData = []
    addNodesData = []
    maxNo = max(list(nodeDict.keys()))
    maxElmNo = max(list(elementDict.keys()))
    for header, data in contHeaderData:
        words = pyFistr.deleteSp(header).split(",")
        if words[0] == "!SPRING":
            ws = pyFistr.deleteSp(data[0]).split(",")
            ngrpName = ws[0]
            spValue = ws[-1]
            ngrpNodes = getNgrpData(meshHeaderData, ngrpName)
            newNodes = []
            elms = []
            for nodeNo in ngrpNodes:
                loc = nodeDict[nodeNo]["loc"]
                for i in range(len(loc)):
                    newLoc = loc[:]
                    newLoc[i] += 1.0
                    maxNo += 1
                    newNodeNo = maxNo
                    node = [newNodeNo] + newLoc
                    newNodes.append(node)
                    maxElmNo += 1
                    elmNo = maxElmNo
                    elm = [elmNo, nodeNo, newNodeNo]
                    elms.append(elm)
            header = "*NODE, NSET=" + "springFix_" + ngrpName + "\n"
            addNodesData.append([header, newNodes])
            elsetName = "elm_" + ngrpName
            elsetHeader = "*ELEMENT, TYPE=SPRINGA, ELSET=" + elsetName + "\n"
            elsetHeaderData.append([elsetHeader, elms])
            spHeader = "*SPRING, ELSET=" + elsetName + "\n"
            spHeaderData.append([spHeader, [spValue + "\n"]])
    return addNodesData + elsetHeaderData + spHeaderData

#
#  getMaterialCard
def getMaterialCard(contHeaderData):
    """ materialCardを取得"""
    addHeaderData = []
    for headerData in contHeaderData:
        header = headerData[0]
        data = headerData[1]
        words = pyFistr.deleteSp(header).split(",")
        if words[0] == "!MATERIAL":
            matName = pyFistr.getValName(header, "NAME")
            newHeader = "*MATERIAL, NAME=" + matName + "\n"
            addHeaderData.append([newHeader, []])
        elif words[0] == "!ELASTIC":
            newHeader = "*ELASTIC\n"
            addHeaderData.append([newHeader, data])
    return addHeaderData

#
#  getBoundaryCard
def getBoundaryCard(contHeaderData, springHeaderData):
    addHeaderData = []
    for header, data in contHeaderData:
        words = pyFistr.deleteSp(header).split(",")
        if words[0] == "!BOUNDARY":
            newHeader = "*BOUNDARY\n"
            addHeaderData.append([newHeader, data])
    if len(springHeaderData) > 0:
        for header, data in springHeaderData:
            words = pyFistr.deleteSp(header).split(",")
            if words[0] == "*NODE":
                ngrpName = pyFistr.getValName(header, "NSET")
                newHeader = "*BOUNDARY\n"
                newData = [ngrpName + ", 1, 1, 0.0\n"]
                newData += [ngrpName + ", 2, 2, 0.0\n"]
                newData += [ngrpName + ", 3, 3, 0.0\n"]
                addHeaderData.append([newHeader, newData])
    return addHeaderData

#
#  getCloadCard
def getCloadCard(contHeaderData):
    addHeaderData = []
    for header, data in contHeaderData:
        words = pyFistr.deleteSp(header).split(",")
        if words[0] == "!CLOAD":
            newHeader = "*CLOAD\n"
            newData = []
            for valLine in data:
                ws = pyFistr.deleteSp(valLine).split(",")
                if float(ws[2]) != 0.0:
                    newData.append(valLine)
            addHeaderData.append([newHeader, newData])
    return addHeaderData

#
#  createLines
def createLines(headerData):
    lines = []
    for header, data in headerData:
        lines.append(header)
        words = pyFistr.deleteSp(header).split(",")
        if words[0] == "*NODE":
            for node in data:
                #座標値を整形する。
                #  整形しないと、実行時エラー発生する場合がある
                strList = [str(node[0])]
                for i in range(1, len(node)):
                    val = "{:16.9e}".format(node[i])
                    strList.append(val)
                line = ", ".join(strList) + "\n"
                lines.append(line)
        elif words[0] == "*ELEMENT":
            for node in data:
                line = ", ".join(list(map(str, node))) + "\n"
                lines.append(line)
        elif words[0] == "*NSET" or words[0] == "*ELSET":
            vals = list(map(str, data))
            for i in range(0, len(vals), 5):
                line = ", ".join(vals[i:i+5]) + "\n"
                lines.append(line)
        else:
            for line in data:
                lines.append(line)
    return lines

#
#  convToInp
#------------
def convToInp(meshHeaderData, contHeaderData):
    """ abaqus形式のinpFileのheaderDataを作成する"""
    #nodeDictを作成
    nodeDict = createNodeDict(meshHeaderData)
    #*NODEを作成
    nodeHeaderData = getNodeCard(meshHeaderData)
    #*ELEMENTを作成
    elmHeaderData, addNodeHeaderData, nodeDict, elementDict = getElementCard(meshHeaderData, nodeDict)
    nodeHeaderData += addNodeHeaderData
    #*NSETを作成
    ngrpHeaderData = getNgrpCard(meshHeaderData)
    #*MATERIALを作成
    matHeaderData = getMaterialCard(contHeaderData)
    #*SECTIONを作成
    sectionHeaderData = getSectionCard(meshHeaderData, elementDict)
    #*EQUATIONを作成
    equationHeaderData = getEquationCard(meshHeaderData)
    #*SPRINGを作成
    springHeaderData = getSpringCard(contHeaderData, meshHeaderData, nodeDict, elementDict)
    #*BOUNDARYを作成
    boundHeaderData = getBoundaryCard(contHeaderData, springHeaderData)
    #*CLOADを作成
    cloadHeaderData = getCloadCard(contHeaderData)
    #全体のheaderDataを作成
    headerData = nodeHeaderData + elmHeaderData + equationHeaderData + ngrpHeaderData
    headerData += springHeaderData + boundHeaderData
    headerData += matHeaderData + sectionHeaderData
    headerData += [["*STEP, INC=100, NLGEOM\n", []]]
    headerData += [["*STATIC\n", []]]
    headerData += cloadHeaderData
    headerData += [["*NODE PRINT, NSET=Nall\n", ["U\n"]]]
    for header, data in elmHeaderData:
        egrp = pyFistr.getValName(header, "ELSET")
        headerData += [["*EL PRINT, ELSET="+egrp+"\n", ["S\n"]]]
    headerData += [["*NODE FILE\n", ["U\n"]]]
    headerData += [["*EL FILE\n", ["S\n"]]]
    headerData += [["*END STEP\n", []]]
    return headerData

#
#  convertToInp
#----------------
def convertToInp(fileHeader):
    """ fistr形式をinp形式に変換する"""
    #fistrのfileを読み込み
    fileName = fileHeader + ".msh"
    f = open(fileName); lines = f.readlines(); f.close()
    meshHeaderData = pyFistr.getHeaderNumData(lines)
    fileName = fileHeader + ".cnt"
    f = open(fileName); lines = f.readlines(); f.close()
    contHeaderData = pyFistr.getHeaderData(lines)
    #inpFileのheaderDataを作成
    abaHeaderData = convToInp(meshHeaderData, contHeaderData)
    #linesを作成し、保存
    lines = createLines(abaHeaderData)
    fileName = fileHeader + ".inp"
    f = open(fileName, "w"); f.writelines(lines); f.close()
    print("FrontISTR形式をinp形式に変換しました。")

#
#  getArgument
#--------------
def getArgument():
    meshFlag = "change2nd"          #1次要素は2次要素に変換する
    fileHeader = "FistrModel"       #FrontISTRのfileHeader名
    if len(sys.argv) == 1:
        return meshFlag, fileHeader
    elif len(sys.argv) == 2:
        if sys.argv[1] == "-n":
            meshFlag = "noChange"
        else:
            fileHeader = sys.argv[1]
        return meshFlag, fileHeader
    else:
        if sys.argv[1] == "-n":
            meshFlag = "noChange"
            fileHeader = sys.argv[2]
        else:
            fileHeader = sys.argv[1]
            if sys.argv[2] == "-n":
                meshFlag = "noChange"
        return meshFlag, fileHeader


if __name__ == "__main__":
    import gettext
    gettext.install("app")
    _ = gettext.gettext

    meshFlag, fileHeader = getArgument()
    convertToInp(fileHeader)
